import zengl


def test_context_info(ctx: zengl.Context):
    assert isinstance(ctx.info, dict)

    assert 'vendor' in ctx.info
    assert 'renderer' in ctx.info
    assert 'version' in ctx.info
    assert 'glsl' in ctx.info

    assert isinstance(ctx.info['vendor'], str)
    assert isinstance(ctx.info['renderer'], str)
    assert isinstance(ctx.info['version'], str)
    assert isinstance(ctx.info['glsl'], str)



import zengl


def make_pipeline(ctx, framebuffer, viewport, color):
    return ctx.pipeline(
        vertex_shader='''
            #version 330 core

            vec2 positions[3] = vec2[](
                vec2(0.1, 0.0),
                vec2(-0.05, 0.086),
                vec2(-0.05, -0.086)
            );

            void main() {
                gl_Position = vec4(positions[gl_VertexID], 0.0, 1.0);
            }
        ''',
        fragment_shader='''
            #version 330 core

            #include "color"

            layout (location = 0) out vec4 out_color;

            void main() {
                out_color = vec4(color, 1.0);
            }
        ''',
        framebuffer=framebuffer,
        viewport=viewport,
        topology='triangles',
        vertex_count=3,
        includes={
            'color': f'vec3 color = vec3({color[0]:.3f}, {color[1]:.3f}, {color[2]:.3f});',
        },
    )


def test(ctx: zengl.Context):
    image = ctx.image((64, 64), 'rgba8unorm')
    depth = ctx.image((64, 64), 'depth24plus')
    depth.clear_value = 0.5

    pipeline_1 = make_pipeline(ctx, [image, depth], (0, 0, 32, 32), (0.0, 0.0, 1.0))
    pipeline_2 = make_pipeline(ctx, [image, depth], (32, 32, 32, 32), (0.0, 0.0, 1.0))
    pipeline_3 = make_pipeline(ctx, [image, depth], (0, 32, 32, 32), (0.0, 1.0, 0.0))
    pipeline_4 = make_pipeline(ctx, [image], (32, 0, 32, 32), (0.0, 0.0, 1.0))

    info_1 = zengl.inspect(pipeline_1)
    info_2 = zengl.inspect(pipeline_2)
    info_3 = zengl.inspect(pipeline_3)
    info_4 = zengl.inspect(pipeline_4)

    assert info_1['vertex_array'] == info_2['vertex_array'] == info_3['vertex_array'] == info_4['vertex_array']

    assert info_1['framebuffer'] == info_2['framebuffer'] == info_3['framebuffer']
    assert info_1['framebuffer'] != info_4['framebuffer']

    assert info_1['program'] == info_2['program'] == info_4['program']
    assert info_1['program'] != info_3['program']



import pytest
import zengl


def test_invalid_buffer_parameters(ctx: zengl.Context):
    with pytest.raises(ValueError):
        ctx.buffer()

    with pytest.raises(TypeError):
        ctx.buffer('test')

    with pytest.raises(TypeError):
        ctx.buffer(100)

    with pytest.raises(ValueError):
        ctx.buffer(b'')

    with pytest.raises(ValueError):
        ctx.buffer(None)

    with pytest.raises(TypeError):
        ctx.buffer(size='999')

    with pytest.raises(ValueError):
        ctx.buffer(size=0)

    with pytest.raises(ValueError):
        ctx.buffer(size=-1)


def test_invalid_buffer_write_parameters(ctx: zengl.Context):
    buf = ctx.buffer(size=32)

    with pytest.raises(TypeError):
        buf.write('test')

    with pytest.raises(TypeError):
        buf.write(10)

    with pytest.raises(TypeError):
        buf.write(None)

    with pytest.raises(ValueError):
        buf.write(b'123', 32 - 2)

    with pytest.raises(ValueError):
        buf.write(b'1', -1)

    with pytest.raises(ValueError):
        buf.write(b'1' * 33)

    with pytest.raises(TypeError):
        buf.write(b'1', offset=0.1)



import numpy as np
import zengl


def test(ctx: zengl.Context):
    image = ctx.image((64, 64), 'rgba8unorm')
    depth = ctx.image((64, 64), 'depth24plus')
    texture = ctx.image((16, 16), 'rgba8unorm', np.full((16, 16, 4), (64, 64, 255, 255), 'u1'))
    uniform_buffer = ctx.buffer(size=64)
    vertex_buffer = ctx.buffer(
        np.array(
            [
                [-1.0, 1.0, -1.0, 1.0],
                [-1.0, -1.0, 1.0, 1.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 1.0],
                [0.0, 0.0, 1.0, 1.0],
            ],
            'f4',
        ).T
    )
    pipeline = ctx.pipeline(
        vertex_shader='''
            #version 330 core

            layout (std140) uniform Common {
                mat4 mvp;
            };

            layout (location = 0) in vec3 in_vert;
            layout (location = 1) in vec2 in_text;

            out vec2 v_text;

            void main() {
                v_text = in_text;
                gl_Position = mvp * vec4(in_vert, 1.0);
            }
        ''',
        fragment_shader='''
            #version 330 core

            uniform sampler2D Texture;

            in vec2 v_text;

            layout (location = 0) out vec4 out_color;

            void main() {
                out_color = texture(Texture, v_text);
            }
        ''',
        layout=[
            {
                'name': 'Common',
                'binding': 0,
            },
            {
                'name': 'Texture',
                'binding': 0,
            },
        ],
        resources=[
            {
                'type': 'uniform_buffer',
                'binding': 0,
                'buffer': uniform_buffer,
            },
            {
                'type': 'sampler',
                'binding': 0,
                'image': texture,
            },
        ],
        framebuffer=[image, depth],
        topology='triangle_strip',
        cull_face='back',
        vertex_buffers=zengl.bind(vertex_buffer, '3f 2f', 0, 1),
        vertex_count=vertex_buffer.size // zengl.calcsize('3f 2f'),
    )

    camera = zengl.camera((0.4, 0.3, 0.2), (0.0, 0.0, 0.0), aspect=1.0, fov=45.0)

    ctx.new_frame()
    uniform_buffer.write(camera)

    image.clear()
    depth.clear()
    pipeline.render()
    image.blit()
    ctx.end_frame()

    pixels = np.frombuffer(image.read(), 'u1').reshape(64, 64, 4)
    np.testing.assert_array_equal(pixels[32, 32], [64, 64, 255, 255])



import pytest
import zengl

simple_vertex_shader = '''
    #version 300 es
    precision highp float;
    void main() {
        gl_Position = vec4(0.0, 0.0, 0.0, 1.0);
    }
'''

simple_fragment_shader = '''
    #version 300 es
    precision highp float;
    layout (location = 0) out vec4 out_color;
    void main() {
        out_color = vec4(0.0);
    }
'''


def test_missing_vertex_shader(ctx: zengl.Context):
    image = ctx.image((4, 4), 'rgba8unorm')
    with pytest.raises(TypeError):
        ctx.pipeline(
            fragment_shader='''
                #version 330 core
                void main() {
                }
            ''',
            framebuffer=[image],
        )


def test_missing_fragment_shader(ctx: zengl.Context):
    image = ctx.image((4, 4), 'rgba8unorm')
    with pytest.raises(TypeError):
        ctx.pipeline(
            vertex_shader='''
                #version 330 core
                void main() {
                    gl_Position = vec4(1.0);
                }
            ''',
            framebuffer=[image],
        )


def test_missing_framebuffer(ctx: zengl.Context):
    with pytest.raises(TypeError):
        ctx.pipeline(
            vertex_shader='''
                #version 330 core
                void main() {
                    gl_Position = vec4(1.0);
                }
            ''',
            fragment_shader='''
                #version 330 core
                void main() {
                }
            ''',
        )


def test_unbound_attribute(ctx: zengl.Context):
    with pytest.raises(ValueError):
        ctx.pipeline(
            vertex_shader='''
                #version 300 es
                precision highp float;
                layout (location = 0) in vec2 in_vertex;
                void main() {
                    gl_Position = vec4(in_vertex, 0.0, 1.0);
                }
            ''',
            fragment_shader=simple_fragment_shader,
            framebuffer=None,
            viewport=(0, 0, 4, 4),
        )


def test_invalid_vertex_format(ctx: zengl.Context):
    vbo = ctx.buffer(size=256)

    with pytest.raises(ValueError):
        ctx.pipeline(
            vertex_shader='''
                #version 300 es
                precision highp float;
                layout (location = 0) in vec2 in_vertex;
                void main() {
                    gl_Position = vec4(in_vertex, 0.0, 1.0);
                }
            ''',
            fragment_shader=simple_fragment_shader,
            framebuffer=None,
            viewport=(0, 0, 4, 4),
            vertex_buffers=[
                {
                    'location': 0,
                    'buffer': vbo,
                    'format': 'float64x2',
                    'offset': 0,
                    'step': 'vertex',
                    'stride': 8,
                },
            ],
        )


def test_invalid_topology(ctx: zengl.Context):
    with pytest.raises(ValueError):
        ctx.pipeline(
            vertex_shader=simple_vertex_shader,
            fragment_shader=simple_fragment_shader,
            framebuffer=None,
            viewport=(0, 0, 4, 4),
            topology='bad',
        )


def test_invalid_render_data(ctx: zengl.Context):
    with pytest.raises(TypeError):
        ctx.pipeline(
            vertex_shader=simple_vertex_shader,
            fragment_shader=simple_fragment_shader,
            framebuffer=None,
            viewport=(0, 0, 4, 4),
            render_data='bad',
        )


def test_invalid_viewport_data(ctx: zengl.Context):
    with pytest.raises(TypeError):
        ctx.pipeline(
            vertex_shader=simple_vertex_shader,
            fragment_shader=simple_fragment_shader,
            framebuffer=None,
            viewport=(0, 0, 4, 4),
            viewport_data=memoryview(bytearray(1000)),
        )


def test_invalid_cull_face(ctx: zengl.Context):
    with pytest.raises(KeyError):
        ctx.pipeline(
            vertex_shader=simple_vertex_shader,
            fragment_shader=simple_fragment_shader,
            framebuffer=None,
            viewport=(0, 0, 4, 4),
            cull_face='bad',
        )



import numpy as np
import zengl


def make_pipeline(ctx, framebuffer, viewport, depth_function):
    return ctx.pipeline(
        vertex_shader='''
            #version 330 core

            vec3 positions[3] = vec3[](
                vec3(0.1, 0.0, 0.0),
                vec3(-0.05, 0.086, 0.0),
                vec3(-0.05, -0.086, 0.0)
            );

            vec3 colors[3] = vec3[](
                vec3(1.0, 0.0, 0.0),
                vec3(0.0, 1.0, 0.0),
                vec3(0.0, 0.0, 1.0)
            );

            out vec3 v_color;

            void main() {
                v_color = colors[gl_InstanceID];
                gl_Position = vec4(positions[gl_VertexID] + vec3(0.0, 0.0, float(gl_InstanceID) - 1.0), 1.0);
            }
        ''',
        fragment_shader='''
            #version 330 core

            in vec3 v_color;

            layout (location = 0) out vec4 out_color;

            void main() {
                out_color = vec4(v_color, 1.0);
            }
        ''',
        framebuffer=framebuffer,
        depth={
            'func': depth_function,
            'write': True,
        },
        viewport=viewport,
        topology='triangles',
        vertex_count=3,
        instance_count=3,
    )


def test(ctx: zengl.Context):
    image = ctx.image((64, 64), 'rgba8unorm')
    depth = ctx.image((64, 64), 'depth24plus')
    depth.clear_value = 0.5

    pipeline_1 = make_pipeline(ctx, [image, depth], (0, 0, 32, 32), 'less')
    pipeline_2 = make_pipeline(ctx, [image, depth], (32, 32, 32, 32), 'always')
    pipeline_3 = make_pipeline(ctx, [image, depth], (0, 32, 32, 32), 'greater')
    pipeline_4 = make_pipeline(ctx, [image, depth], (32, 0, 32, 32), 'equal')

    ctx.new_frame()
    image.clear()
    depth.clear()
    pipeline_1.render()
    pipeline_2.render()
    pipeline_3.render()
    pipeline_4.render()
    ctx.end_frame()
    pixels = np.frombuffer(image.read(), 'u1').reshape(64, 64, 4)
    np.testing.assert_array_equal(
        pixels[[16, 16, 48, 48], [16, 48, 16, 48]],
        [
            [255, 0, 0, 255],
            [0, 255, 0, 255],
            [0, 0, 255, 255],
            [0, 0, 255, 255],
        ],
    )



import zengl
import numpy as np


def make_pipeline(ctx, image, gltype, color):
    return ctx.pipeline(
        vertex_shader='''
            #version 300 es
            precision highp float;
            vec2 positions[3] = vec2[](vec2(-1.0, -1.0), vec2(3.0, -1.0), vec2(-1.0, 3.0));
            void main() {
                gl_Position = vec4(positions[gl_VertexID], 0.0, 1.0);
            }
        ''',
        fragment_shader='''
            #version 300 es
            precision highp float;
            precision highp int;
            #include "TYPE"
            uniform TYPE color;
            layout (location = 0) out TYPE out_color;
            void main() {
                out_color = color;
            }
        ''',
        includes={
            'TYPE': f'#define TYPE {gltype}',
        },
        uniforms={
            'color': color,
        },
        framebuffer=[image],
        topology='triangles',
        vertex_count=3,
    )


def test_uniform_vec4(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba32float', bytearray(256))

    pipeline = make_pipeline(ctx, img, 'vec4', [0.0, 0.25, 0.5, 1.0])
    pipeline.render()
    pixel = np.frombuffer(img.read((1, 1)), 'f4')
    np.testing.assert_array_almost_equal(pixel, [0.0, 0.25, 0.5, 1.0], decimal=3)

    pipeline = make_pipeline(ctx, img, 'vec4', [0.33, 0.33, 0.33, 1.0])
    pipeline.render()
    pixel = np.frombuffer(img.read((1, 1)), 'f4')
    np.testing.assert_array_almost_equal(pixel, [0.33, 0.33, 0.33, 1.0], decimal=3)


def test_uniform_ivec4(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba32sint', bytearray(256))

    pipeline = make_pipeline(ctx, img, 'ivec4', [0, 100, -1024, 0x11223344])
    pipeline.render()
    pixel = np.frombuffer(img.read((1, 1)), 'i4')
    np.testing.assert_array_equal(pixel, [0, 100, -1024, 0x11223344])

    pipeline = make_pipeline(ctx, img, 'ivec4', [-1, -1, 100000, 100000])
    pipeline.render()
    pixel = np.frombuffer(img.read((1, 1)), 'i4')
    np.testing.assert_array_equal(pixel, [-1, -1, 100000, 100000])


def test_uniform_uvec4(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba32uint', bytearray(256))

    pipeline = make_pipeline(ctx, img, 'uvec4', [0, 100, 1024, 0xFF000000])
    pipeline.render()
    pixel = np.frombuffer(img.read((1, 1)), 'u4')
    np.testing.assert_array_equal(pixel, [0, 100, 1024, 0xFF000000])

    pipeline = make_pipeline(ctx, img, 'uvec4', [255, 127, 63, 0])
    pipeline.render()
    pixel = np.frombuffer(img.read((1, 1)), 'u4')
    np.testing.assert_array_equal(pixel, [255, 127, 63, 0])


def test_uniform_mat4(ctx: zengl.Context):
    img = ctx.image((4, 4), 'rgba32float', bytearray(256))

    pipeline = ctx.pipeline(
        vertex_shader='''
            #version 300 es
            precision highp float;
            vec2 positions[3] = vec2[](vec2(-1.0, -1.0), vec2(3.0, -1.0), vec2(-1.0, 3.0));
            void main() {
                gl_Position = vec4(positions[gl_VertexID], 0.0, 1.0);
            }
        ''',
        fragment_shader='''
            #version 300 es
            precision highp float;
            uniform mat4 kernel;
            uniform vec4 color;
            layout (location = 0) out vec4 out_color;
            void main() {
                out_color = kernel * color;
            }
        ''',
        uniforms={
            'kernel': [0.0] * 16,
            'color': [1.0, 2.0, 3.0, 4.0],
        },
        framebuffer=[img],
        topology='triangles',
        vertex_count=3,
    )

    kernel = [
        [1.0, 0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0, 0.0],
        [0.0, 0.0, 1.0, 0.0],
        [0.0, 0.0, 0.0, 1.0],
    ]

    pipeline.uniforms['kernel'][:] = np.array(kernel, 'f4').tobytes()
    pipeline.render()
    pixel = np.frombuffer(img.read((1, 1)), 'f4')
    np.testing.assert_array_almost_equal(pixel, [1.0, 2.0, 3.0, 4.0], decimal=3)

    kernel = [
        [-2.053, 0.115, 0.683, -0.524],
        [-1.753, -1.090, -0.333, -0.641],
        [1.937, 0.477, -3.711, -0.760],
        [-0.865, 1.234, 2.012, 0.132],
    ]

    pipeline.uniforms['kernel'][:] = np.array(kernel, 'f4').tobytes()
    pipeline.render()
    pixel = np.frombuffer(img.read((1, 1)), 'f4')
    np.testing.assert_array_almost_equal(pixel, [-3.208, 4.302, -3.068, -3.558], decimal=3)



import pytest
import zengl


def test_buffer_data(ctx: zengl.Context):
    buf = ctx.buffer(b'Hello World!')
    assert buf.read() == b'Hello World!'


def test_buffer_data_non_contiguous(ctx: zengl.Context):
    buf = ctx.buffer(memoryview(b'Hello World!')[::2])
    assert buf.read() == b'HloWrd'


def test_buffer_write(ctx: zengl.Context):
    buf = ctx.buffer(size=12)
    buf.write(b'Hello World!')
    assert buf.read() == b'Hello World!'


def test_buffer_write_offset(ctx: zengl.Context):
    buf = ctx.buffer(b'---------------')
    buf.write(b'Hello', offset=2)
    buf.write(b'World', offset=9)
    assert buf.read() == b'--Hello--World-'


def test_buffer_copy(ctx: zengl.Context):
    buf1 = ctx.buffer(b'aaaaaaaaaa')
    buf2 = ctx.buffer(b'bbbbbbbbbb')
    assert buf2.read() == b'bbbbbbbbbb'
    buf2.write(buf1)
    assert buf2.read() == b'aaaaaaaaaa'


def test_buffer_copy_with_read(ctx: zengl.Context):
    buf1 = ctx.buffer(b'aaaaaaaaaa')
    buf2 = ctx.buffer(b'bbbbbbbbbb')
    assert buf2.read() == b'bbbbbbbbbb'
    buf1.read(into=buf2)
    assert buf2.read() == b'aaaaaaaaaa'


def test_buffer_copy_offset(ctx: zengl.Context):
    buf1 = ctx.buffer(b'aaaaaaaaaa')
    buf2 = ctx.buffer(b'bbbbbbbbbb')
    assert buf2.read() == b'bbbbbbbbbb'
    buf2.write(buf1.view(4), offset=4)
    assert buf2.read() == b'bbbbaaaabb'


def test_buffer_copy_offset_with_read(ctx: zengl.Context):
    buf1 = ctx.buffer(b'aaaaaaaaaa')
    buf2 = ctx.buffer(b'bbbbbbbbbb')
    assert buf2.read() == b'bbbbbbbbbb'
    buf1.read(size=4, into=buf2.view(offset=4, size=4))
    assert buf2.read() == b'bbbbaaaabb'


def test_buffer_write_error(ctx: zengl.Context):
    buf = ctx.buffer(b'bbbbbbbbbb')

    with pytest.raises(ValueError):
        buf.write(b'aaaaaaaaaa', offset=-1)

    with pytest.raises(ValueError):
        buf.write(b'aaaaaaaaaa', offset=4)


def test_buffer_copy_error(ctx: zengl.Context):
    buf1 = ctx.buffer(b'aaaaaaaaaa')
    buf2 = ctx.buffer(b'bbbbbbbbbb')

    with pytest.raises(ValueError):
        buf2.write(buf1.view(), offset=-1)

    with pytest.raises(ValueError):
        buf2.write(buf1.view(), offset=4)


def test_buffer_data_and_size(ctx: zengl.Context):
    with pytest.raises(ValueError):
        ctx.buffer(b'Hello World!', size=64)


def test_buffer_read_offset(ctx: zengl.Context):
    buf = ctx.buffer(b'0123456789')
    assert buf.read() == b'0123456789'
    assert buf.read(4) == b'0123'
    assert buf.read(offset=5) == b'56789'
    assert buf.read(offset=6, size=3) == b'678'


def test_buffer_read_error(ctx: zengl.Context):
    buf = ctx.buffer(b'bbbbbbbbbb')

    with pytest.raises(ValueError):
        buf.read(100)

    with pytest.raises(ValueError):
        buf.read(offset=-1)

    with pytest.raises(ValueError):
        buf.read(offset=4, size=10)

    with pytest.raises(ValueError):
        buf.read(size=-1)


def test_buffer_read_into(ctx: zengl.Context):
    buf = ctx.buffer(b'0123456789')

    mem = memoryview(bytearray(10))
    buf.read(into=mem)
    assert bytes(mem) == b'0123456789'

    mem = memoryview(bytearray(4))
    buf.read(4, into=mem)
    assert bytes(mem) == b'0123'

    mem = memoryview(bytearray(5))
    buf.read(offset=5, into=mem)
    assert bytes(mem) == b'56789'

    mem = memoryview(bytearray(3))
    buf.read(offset=6, size=3, into=mem)
    assert bytes(mem) == b'678'


def test_invalid_access(ctx):
    with pytest.raises(ValueError):
        ctx.buffer(size=64, access='bad')



import numpy as np
import zengl


def test_read_depth16(ctx: zengl.Context):
    img = ctx.image((4, 4), 'depth16unorm')

    img.clear_value = 0.0
    img.clear()
    content = np.frombuffer(img.read(), 'u2').reshape(4, 4).astype(float) / 0xFFFF
    expected = np.zeros((4, 4))
    np.testing.assert_array_almost_equal(content, expected, 2)

    img.clear_value = 0.5
    img.clear()
    content = np.frombuffer(img.read(), 'u2').reshape(4, 4).astype(float) / 0xFFFF
    expected = np.full((4, 4), 0.5)
    np.testing.assert_array_almost_equal(content, expected, 2)


def test_read_depth24(ctx: zengl.Context):
    img = ctx.image((4, 4), 'depth24plus')

    img.clear_value = 0.0
    img.clear()
    packed = np.frombuffer(img.read(), 'u4').reshape(4, 4)
    content = (packed >> 8).astype(float) / 0xFFFFFF
    expected = np.zeros((4, 4))
    np.testing.assert_array_almost_equal(content, expected, 2)

    img.clear_value = 0.5
    img.clear()
    packed = np.frombuffer(img.read(), 'u4').reshape(4, 4)
    content = (packed >> 8).astype(float) / 0xFFFFFF
    expected = np.full((4, 4), 0.5)
    np.testing.assert_array_almost_equal(content, expected, 2)


def test_read_depth24_stencil8(ctx: zengl.Context):
    img = ctx.image((4, 4), 'depth24plus-stencil8')

    img.clear_value = (0.0, 0)
    img.clear()
    packed = np.frombuffer(img.read(), 'u4').reshape(4, 4)
    content = (packed >> 8).astype(float) / 0xFFFFFF
    expected = np.zeros((4, 4))
    np.testing.assert_array_almost_equal(content, expected, 2)
    content = packed & 0xFF
    expected = np.zeros((4, 4), int)
    np.testing.assert_array_equal(content, expected)

    img.clear_value = (0.5, 17)
    img.clear()
    packed = np.frombuffer(img.read(), 'u4').reshape(4, 4)
    content = (packed >> 8).astype(float) / 0xFFFFFF
    expected = np.full((4, 4), 0.5)
    np.testing.assert_array_almost_equal(content, expected, 2)
    content = packed & 0xFF
    expected = np.full((4, 4), 17)
    np.testing.assert_array_equal(content, expected)


def test_read_depth32(ctx: zengl.Context):
    img = ctx.image((4, 4), 'depth32float')

    img.clear_value = 0.0
    img.clear()
    content = np.frombuffer(img.read(), 'f4').reshape(4, 4)
    expected = np.zeros((4, 4))
    np.testing.assert_array_almost_equal(content, expected, 2)

    img.clear_value = 0.5
    img.clear()
    content = np.frombuffer(img.read(), 'f4').reshape(4, 4)
    expected = np.full((4, 4), 0.5)
    np.testing.assert_array_almost_equal(content, expected, 2)



import struct

import numpy as np
import zengl


def test(ctx: zengl.Context):
    image = ctx.image((64, 64), 'rgba8unorm')
    temp = np.concatenate(
        [
            np.full((16, 16, 2), (-0.5, -0.5)),
            np.full((16, 16, 2), (0.5, -0.5)),
            np.full((16, 16, 2), (-0.5, 0.5)),
            np.full((16, 16, 2), (0.5, 0.5)),
            np.full((16, 16, 2), (0.0, 0.0)),
            np.full((16, 16, 2), (0.0, 0.0)),
        ],
        dtype='f4',
    )
    texture = ctx.image((16, 16), 'rg32float', temp, cubemap=True)
    pipeline = ctx.pipeline(
        vertex_shader='''
            #version 330 core

            uniform samplerCube Texture;

            vec2 positions[3] = vec2[](
                vec2(0.1, 0.0),
                vec2(-0.05, 0.086),
                vec2(-0.05, -0.086)
            );

            uniform vec3 ray;

            void main() {
                gl_Position = vec4(positions[gl_VertexID] + texture(Texture, ray).rg, 0.0, 1.0);
            }
        ''',
        fragment_shader='''
            #version 330 core

            layout (location = 0) out vec4 out_color;

            void main() {
                out_color = vec4(0.0, 0.0, 1.0, 1.0);
            }
        ''',
        layout=[
            {
                'name': 'Texture',
                'binding': 5,
            },
        ],
        resources=[
            {
                'type': 'sampler',
                'binding': 5,
                'image': texture,
            },
        ],
        uniforms={
            'ray': (0.0, 0.0, 0.0),
        },
        framebuffer=[image],
        topology='triangles',
        vertex_count=3,
    )

    ctx.new_frame()
    image.clear()
    struct.pack_into('3f', pipeline.uniforms['ray'], 0, 0.9, 0.1, 0.2)
    pipeline.render()
    struct.pack_into('3f', pipeline.uniforms['ray'], 0, -0.9, 0.2, 0.1)
    pipeline.render()
    struct.pack_into('3f', pipeline.uniforms['ray'], 0, 0.1, 0.9, -0.2)
    pipeline.render()
    struct.pack_into('3f', pipeline.uniforms['ray'], 0, -0.2, -0.9, 0.1)
    pipeline.render()
    ctx.end_frame()
    pixels = np.frombuffer(image.read(), 'u1').reshape(64, 64, 4)
    np.testing.assert_array_equal(
        pixels[[16, 16, 48, 48], [16, 48, 16, 48]],
        [
            [0, 0, 255, 255],
            [0, 0, 255, 255],
            [0, 0, 255, 255],
            [0, 0, 255, 255],
        ],
    )
